import math

import scipy
import sklearn


class DiversityMetric:
    """
   Abstract Class for ensemble metrics.
   """

    def get_name(self):
        """
        Returns the name of the metric (as string)
        """
        pass

    def compute_diversity(self, pred1, pred2, test_y):
        """
        Abstract method to compute diversity metric
        :param pred1: predictions of the first classifier
        :param pred2: predictions of the second classifier
        :param test_y: reference test labels
        :return: a float value
        """
        pass

    def compute_n(self, pred1, pred2, test_y, rel=False):
        """
        Method to compute n00, n01, n10, n11
        :param pred1: predictions of the first classifier
        :param pred2: predictions of the second classifier
        :param test_y: reference test labels
        :return: n00, n01, n10, n11
        """
        n11 = sum((pred1 == test_y) * (pred2 == test_y))
        n10 = sum((pred1 == test_y) * (pred2 != test_y))
        n01 = sum((pred1 != test_y) * (pred2 == test_y))
        n00 = sum((pred1 != test_y) * (pred2 != test_y))
        if rel:
            n11 = 1.0*n11/len(test_y)
            n10 = 1.0*n10/len(test_y)
            n01 = 1.0*n01/len(test_y)
            n00 = 1.0*n00/len(test_y)
        return n00, n01, n10, n11


class RSquaredMetric(DiversityMetric):
    """
    Ranker using the R-Squared Statistical Index from SciPy
    """

    def compute_diversity(self, pred1, pred2, test_y):
        slope, intercept, r_value, p_value, std_err = scipy.stats.linregress(pred1, pred2)
        return r_value ** 2

    def get_name(self):
        return "R-Squared"


class PearsonMetric(DiversityMetric):
    """
    Ranker using the Pearson Correlation Index from SciPy
    """

    def compute_diversity(self, pred1, pred2, test_y):
        return scipy.stats.pearsonr(pred1, pred2)[0]

    def get_name(self):
        return "Pearson"


class CosineSimilarityMetric(DiversityMetric):
    """
    Ranker using the Cosine Distance from SciPy
    """

    def compute_diversity(self, pred1, pred2, test_y):
        return 1 - scipy.spatial.distance.cosine(pred1, pred2)

    def get_name(self):
        return "Cosine"


class SpearmanMetric(DiversityMetric):
    """
    Ranker using the Spearman Statistical Index from SciPy
    """

    def compute_diversity(self, pred1, pred2, test_y):
        return scipy.stats.spearmanr(pred1, pred2)[0]

    def get_name(self):
        return "Spearman"


class ChiSquaredMetric(DiversityMetric):
    """
    Ranker using the Chi-Squared Statistical Index from Scikit-Learn
    """

    def compute_diversity(self, pred1, pred2, test_y):
        scaled_feat_values = sklearn.preprocessing.MinMaxScaler().fit_transform(pred1.reshape(-1, 1))
        return sklearn.feature_selection.chi2(scaled_feat_values, pred2)[0][0]

    def get_name(self):
        return "ChiSquared"


class MutualInfoMetric(DiversityMetric):
    """
    Ranker using the Mutual Information Index from Scikit-Learn
    """

    def compute_diversity(self, pred1, pred2, test_y):
        return sklearn.feature_selection.mutual_info_classif(pred1.reshape(-1, 1), pred2)[0]

    def get_name(self):
        return "MutualInfo"


class ANOVAMetric(DiversityMetric):
    """
    Ranker using the ANOVA R from Scikit-Learn
    """

    def compute_diversity(self, pred1, pred2, test_y):
        return sklearn.feature_selection.f_classif(pred1.reshape(-1, 1), pred2)[0]

    def get_name(self):
        return "ANOVAF"


class QStatDiversity(DiversityMetric):
    """
    Metric using the QStat Diversity
    """

    def compute_diversity(self, pred1, pred2, test_y):
        n00, n01, n10, n11 = self.compute_n(pred1, pred2, test_y)
        return (n11*n00 - n01*n10)/(n11*n00 + n01*n10)

    def get_name(self):
        return "QStat"


class SigmaDiversity(DiversityMetric):
    """
    Metric using the Sigma Diversity
    """

    def compute_diversity(self, pred1, pred2, test_y):
        n00, n01, n10, n11 = self.compute_n(pred1, pred2, test_y, rel=True)
        return (n11*n00 - n01*n10)/math.sqrt((n11+n10)*(n11+n01)*(n00+n10)*(n00+n01))

    def get_name(self):
        return "Sigma"


class Disagreement(DiversityMetric):
    """
    Metric using the Disagreement Diversity
    """

    def compute_diversity(self, pred1, pred2, test_y):
        n00, n01, n10, n11 = self.compute_n(pred1, pred2, test_y)
        return (n01 + n10)/(n00 + n01 + n10 + n11)

    def get_name(self):
        return "Disagreement"


class DoubleFault(DiversityMetric):
    """
    Metric using the DoubleFault Diversity
    """

    def compute_diversity(self, pred1, pred2, test_y):
        n00, n01, n10, n11 = self.compute_n(pred1, pred2, test_y)
        return n00/(n00 + n01 + n10 + n11)

    def get_name(self):
        return "DoubleFault"



import copy

import numpy
from pyod.models.base import BaseDetector
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.preprocessing import LabelEncoder
from sklearn.utils import check_X_y
from sklearn.utils.multiclass import unique_labels
from sklearn.utils.validation import check_is_fitted, check_array
from xgboost import XGBClassifier

# ---------------------------------- SUPPORT METHODS ------------------------------------
from confens.metrics.EnsembleMetric import get_default


def get_feature_importance(clf):
    if isinstance(clf, Classifier):
        return clf.feature_importances()
    else:
        return clf.feature_importances_


def auto_bag_rate(n_features):
    """
    Method used to automatically devise a rate to include features in bagging
    :param n_features: number of features in the dataset
    :return: the rate of features to bag
    """
    if n_features < 20:
        bag_rate = 0.8
    elif n_features < 50:
        bag_rate = 0.7
    elif n_features < 100:
        bag_rate = 0.6
    else:
        bag_rate = 0.5
    return bag_rate


def predict_uns_proba(uns_clf, test_features):
    proba = uns_clf.predict_proba(test_features)
    pred = numpy.argmax(proba, axis=1)
    for i in range(len(pred)):
        min_p = min(proba[i])
        max_p = max(proba[i])
        proba[i][pred[i]] = max_p
        proba[i][1 - pred[i]] = min_p
    return proba


class Classifier(ClassifierMixin, BaseEstimator):
    """
    Basic Abstract Class for Classifiers.
    Abstract methods are only the classifier_name, with many degrees of freedom in implementing them.
    Wraps implementations from different frameworks (if needed), sklearn and many deep learning utilities
    """

    def __init__(self, clf):
        """
        Constructor of a generic Classifier
        :param clf: algorithm to be used as Classifier
        """
        super().__init__()
        self.clf = copy.deepcopy(clf) if clf is not None else None
        self._estimator_type = "classifier"
        self.feature_importances_ = None
        self.X_ = None
        self.y_ = None

    def fit(self, X, y=None):

        # Check that X and y have correct shape
        X, y = check_X_y(X, y)

        # Store the classes seen during fit + other data
        if y is not None:
            self.classes_ = unique_labels(y)
        else:
            self.classes_ = [0, 1]
        # self.X_ = X
        # self.y_ = y

        # Train clf
        self.clf.fit(X, y)
        self.feature_importances_ = self.compute_feature_importances()

        # Return the classifier
        return self

    def predict(self, X):
        """
        Method to compute predict of a classifier
        :return: array of predicted class
        """
        probas = self.predict_proba(X)
        if self.is_unsupervised():
            return numpy.argmax(probas, axis=1)
        else:
            return self.classes_[numpy.argmax(probas, axis=1)]

    def predict_proba(self, X):
        """
        Method to compute probabilities of predicted classes
        :return: array of probabilities for each classes
        """

        # Check if fit has been called
        check_is_fitted(self)
        X = check_array(X)

        return self.clf.predict_proba(X)

    def decision_function(self, X):
        """
        Compatibility with PYOD
        :param X: the test set
        :return: a numpy array, or None
        """
        if X is None:
            return None
        X = check_array(X)
        probas = self.predict_proba(X)
        if probas.shape[1] >= 2:
            a = probas[:, 1] / probas[:, 0]
            return a
        else:
            return numpy.zeros(X.shape[0])

    def predict_confidence(self, X):
        """
        Method to compute confidence in the predicted class
        :return: max probability as default
        """
        probas = self.predict_proba(X)
        return numpy.max(probas, axis=1)

    def compute_feature_importances(self):
        """
        Outputs feature ranking in building a Classifier
        :return: ndarray containing feature ranks
        """
        if hasattr(self.clf, 'feature_importances_'):
            return self.clf.feature_importances_
        elif hasattr(self.clf, 'coef_'):
            return numpy.sum(numpy.absolute(self.clf.coef_), axis=0)
        return []

    def is_unsupervised(self):
        """
        true if the classifier is unsupervised
        :return: boolean
        """
        return hasattr(self, 'classes_') and numpy.array_equal(self.classes_, [0, 1])

    def classifier_name(self):
        """
        Returns the name of the classifier (as string)
        """
        return self.clf.__class__.__name__

    def get_diversity(self, X, y, metrics=None):
        """
        Returns diversity metrics. Works only with ensembles.
        :param metrics: name of the metrics to output (list of Metric objects)
        :param X: test set
        :param y: labels of the test set
        :return: diversity metrics
        """
        X = check_array(X)
        predictions = []
        check_is_fitted(self)
        if hasattr(self, "estimators_"):
            # If it is an ensemble and if it is trained
            for baselearner in self.estimators_:
                predictions.append(baselearner.predict(X))
            predictions = numpy.column_stack(predictions)
        elif self.clf is not None:
            # If it wraps an ensemble
            check_is_fitted(self.clf)
            if hasattr(self.clf, "estimators_"):
                # If it is an ensemble and if it is trained
                for baselearner in self.clf.estimators_:
                    predictions.append(baselearner.predict(X))
                predictions = numpy.column_stack(predictions)
        if predictions is not None and len(predictions) > 0:
            # Compute metrics
            metric_scores = {}
            if metrics is None or not isinstance(metrics, list):
                metrics = get_default()
            for metric in metrics:
                metric_scores[metric.get_name()] = metric.compute_diversity(predictions, y)
            return metric_scores
        else:
            # If it is not an ensemble
            return {}

    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            setattr(self.clf, parameter, value)
        return self


class UnsupervisedClassifier(Classifier, BaseDetector):
    """
    Wrapper for unsupervised classifiers belonging to the library PYOD
    """

    def __init__(self, clf):
        """
        Constructor of a generic UnsupervisedClassifier. Assumes that clf is an algorithm from pyod
        :param clf: pyod algorithm to be used as Classifier
        """
        self.clf = clf
        self.contamination = clf.contamination
        self._estimator_type = "classifier"
        self.feature_importances_ = None
        self.X_ = None
        self.y_ = None

    def fit(self, X, y=None):
        # Store the classes seen during fit + other data
        self.classes_ = [0, 1]
        self.X_ = X
        self.y_ = None

        # Train clf
        self.clf.fit(X)
        self.feature_importances_ = self.compute_feature_importances()

        # Return the classifier
        return self

    def decision_function(self, X):
        """
        pyod function to override. Calls the wrapped classifier.
        :param X: test set
        :return: decision function
        """
        return self.clf.decision_function(X)

    def predict_proba(self, X):
        """
        Method to compute probabilities of predicted classes.
        It has to e overridden since PYOD's implementation of predict_proba is wrong
        :return: array of probabilities for each classes
        """

        # Check if fit has been called
        check_is_fitted(self)
        X = check_array(X)

        pred_score = self.decision_function(X)
        probs = numpy.zeros((X.shape[0], 2))
        if isinstance(self.contamination, (float, int)):
            pred_thr = pred_score - self.clf.threshold_
        min_pt = min(pred_thr)
        max_pt = max(pred_thr)
        anomaly = pred_thr > 0
        cont = numpy.asarray([pred_thr[i] / max_pt if anomaly[i] else (pred_thr[i] / min_pt if min_pt != 0 else 0.2)
                              for i in range(0, len(pred_thr))])
        probs[:, 0] = 0.5 + cont / 2
        probs[:, 1] = 1 - probs[:, 0]
        probs[anomaly, 0], probs[anomaly, 1] = probs[anomaly, 1], probs[anomaly, 0]
        return probs

    def classifier_name(self):
        """
        Returns the name of the classifier (as string)
        """
        return self.clf.__class__.__name__


class XGB(Classifier):
    """
    Wrapper for the xgboost.XGBClassifier algorithm
    Can be used as reference to see what's needed to extend the Classifier class
    """

    def __init__(self, n_estimators=100):
        Classifier.__init__(self, XGBClassifier(n_estimators=n_estimators))
        self.l_encoder = None

    def fit(self, X, y=None):
        # Check that X and y have correct shape
        X, y = check_X_y(X, y)

        # Store the classes seen during fit + other data
        self.classes_ = unique_labels(y)
        self.l_encoder = LabelEncoder()
        y = self.l_encoder.fit_transform(y)

        # self.X_ = X
        # self.y_ = y

        # Train clf
        self.clf.fit(X, y)
        self.feature_importances_ = self.compute_feature_importances()

        # Return the classifier
        return self

    def classifier_name(self):
        return "XGBClassifier(" + str(self.clf.n_estimators) + ")"



import copy
import random
from multiprocessing import Pool, Queue
from multiprocessing.pool import ThreadPool

from confens.classifiers.ConfidenceEnsemble import ConfidenceEnsemble
from confens.utils.classifier_utils import get_classifier_name
from confens.utils.general_utils import current_ms


class ConfidenceBagging(ConfidenceEnsemble):
    """
    Class for creating bagging ensembles
    """

    def __init__(self, clf, n_base: int = 10, max_features: float = 0.7, sampling_ratio: float = 0.7,
                 conf_thr: float = None, perc_decisors: float = None, n_decisors: int = None,
                 weighted: bool = False, parallel_train: bool = True):
        """
        Constructor
        :param clf: the algorithm to be used for creating base learners
        :param n_base: number of base learners (= size of the ensemble)
        :param max_features: percentage of features to be used at each iteration
        :param sampling_ratio: percentage of the dataset to be used at each iteration
        :param conf_thr: float value for confidence threshold
        :param perc_decisors: percentage of base learners to be used for prediction
        :param n_decisors: number of base learners to be used for prediction
        :param weighted: True if prediction has to be computed as a weighted sum of probabilities
        """
        super().__init__(clf, n_base, conf_thr, perc_decisors, n_decisors, weighted)
        self.max_features = max_features if max_features is not None and 0 < max_features <= 1 else 0.7
        self.sampling_ratio = sampling_ratio if sampling_ratio is not None and 0 < sampling_ratio <= 1 else 0.7
        self.feature_sets = []
        self.parallel_train = parallel_train

    def train_base_bagger(self, X, y, samples_n, features, learner_index):
        sample_x, sample_y = self.draw_samples(X, y, samples_n)
        sample_x = sample_x[:, features]
        if len(features) == 1:
            sample_x = sample_x.reshape(-1, 1)
        # Train learner
        learner = copy.deepcopy(self.clf_list[learner_index % len(self.clf_list)])
        learner.fit(sample_x, sample_y)
        if hasattr(learner, "X_"):
            learner.X_ = None
        if hasattr(learner, "y_"):
            learner.y_ = None
        # Test Learner
        return learner

    def fit_ensemble(self, X, y=None):
        train_n = len(X)
        bag_features_n = int(X.shape[1] * self.max_features)
        samples_n = int(train_n * self.sampling_ratio)
        # Drawing features
        for learner_index in range(0, self.n_base):
            features = random.sample(range(X.shape[1]), bag_features_n)
            features.sort()
            self.feature_sets.append(features)
        if self.parallel_train:
            # Training in parallel
            with ThreadPool(self.n_base) as p:
                self.estimators_ = p.starmap(self.train_base_bagger,
                                             [(X, y, samples_n, self.feature_sets[i], i) for i in range(0, self.n_base)])
        else:
            self.estimators_ = []
            for learner_index in range(0, self.n_base):
                self.estimators_.append(self.train_base_bagger(X, y, samples_n, self.feature_sets[learner_index], learner_index))

    def classifier_name(self):
        """
        Gets classifier name as string
        :return: the classifier name
        """
        clf_name = get_classifier_name(self.clf)
        if self.weighted:
            return "ConfidenceBaggerWeighted(" + str(clf_name) + "-" + str(self.n_base) + "-" + \
                   str(self.conf_thr) + "-" + str(self.perc_decisors) + "-" + str(self.n_decisors) + "-" + \
                   str(self.max_features) + "-" + str(self.sampling_ratio) + ")"
        else:
            return "ConfidenceBagger(" + str(clf_name) + "-" + str(self.n_base) + "-" + \
                   str(self.conf_thr) + "-" + str(self.perc_decisors) + "-" + str(self.n_decisors) + "-" + \
                   str(self.max_features) + "-" + str(self.sampling_ratio) + ")"



# Support libs
import random

import numpy
import pandas
import sklearn.metrics as metrics
import sklearn.model_selection as ms
# Used to save a classifier and measure its size in KB
from sklearn.calibration import CalibratedClassifierCV
# Name of the folder in which look for tabular (CSV) datasets
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import TunedThresholdClassifierCV, GridSearchCV
from sklearn.tree import DecisionTreeClassifier

from confens.classifiers.ConfidenceBagging import ConfidenceBagging
from confens.classifiers.ConfidenceBoosting import ConfidenceBoosting
# Scikit-Learn algorithms
# The PYOD library contains implementations of unsupervised classifiers.
# Works only with anomaly detection (no multi-class)
# ------- GLOBAL VARS -----------
from confens.utils.classifier_utils import get_classifier_name
from confens.utils.general_utils import current_ms

CSV_FILE = "sample_data/sample_data_arancino.csv"
# Name of the column that contains the label in the tabular (CSV) dataset
LABEL_NAME = 'multilabel'
# Percantage of test data wrt train data
TT_SPLIT = 0.5
# True if debug information needs to be shown
VERBOSE = True

FORCE_BINARY = True

# Set random seed for reproducibility
random.seed(42)
numpy.random.seed(42)

# ----------------------- MAIN ROUTINE ---------------------
# This files shows an example of confidence ensembles for supervised learning.
# Can be used as is, just change the details to load the dataset

def get_base():
    return RandomForestClassifier(n_estimators=10)

if __name__ == '__main__':

    # if file is a CSV, it is assumed to be a dataset to be processed
    df = pandas.read_csv(CSV_FILE, sep=",", nrows=10000)
    if VERBOSE:
        print("\n------------ DATASET INFO -----------------")
        print("Data Points in Dataset '%s': %d" % (CSV_FILE, len(df.index)))
        print("Features in Dataset: " + str(len(df.columns)))

    # Set up train test split excluding categorical values that some algorithms cannot handle
    # 1-Hot-Encoding or other approaches may be used instead of removing
    y = df[LABEL_NAME].to_numpy()
    if FORCE_BINARY:
        y = numpy.where(df[LABEL_NAME] == "normal", 0, 1)
    x_train, x_test, y_train, y_test = ms.train_test_split(df.drop(columns=[LABEL_NAME]).to_numpy(), y,
                                                           test_size=TT_SPLIT, shuffle=True)
    # Creating classifiers
    cboost_parameters = {'n_base': [5, 10, 20], 'perc_decisors': [0.3, 0.5, 0.8, 1.0]}
    classifiers = [
        get_base(),
        TunedThresholdClassifierCV(estimator=get_base()),
        ConfidenceBagging(clf=get_base()),
        ConfidenceBoosting(clf=get_base()),
        TunedThresholdClassifierCV(estimator=ConfidenceBoosting(clf=get_base())),
        CalibratedClassifierCV(estimator=ConfidenceBoosting(clf=get_base()), method='sigmoid'),
        CalibratedClassifierCV(estimator=ConfidenceBoosting(clf=get_base()), method='isotonic'),
        GridSearchCV(estimator=ConfidenceBoosting(clf=get_base()), param_grid=cboost_parameters)
    ]

    for classifier in classifiers:
        # Exercising Classifier
        clf_name = get_classifier_name(classifier)
        start_time = current_ms()
        classifier.fit(x_train, y_train)
        end_time = current_ms()
        clf_pred = classifier.predict(x_test)

        print('%s has accuracy of %.3f, training in %d ms' %
              (clf_name, metrics.accuracy_score(y_test, clf_pred), end_time-start_time))  # , metrics.accuracy_score(y_test, cb_pred)))



# Support libs
import os
import random
import time

import numpy
import pandas
import sklearn.metrics as metrics
import sklearn.model_selection as ms
# Used to save a classifier and measure its size in KB
from joblib import dump
from logitboost import LogitBoost
from pyod.models.cblof import CBLOF
from pyod.models.hbos import HBOS
from pyod.models.iforest import IForest
from pyod.models.inne import INNE
from pyod.models.pca import PCA
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
# Scikit-Learn algorithms
from sklearn.ensemble import RandomForestClassifier, ExtraTreesClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler
from sklearn.tree import DecisionTreeClassifier

# Name of the folder in which look for tabular (CSV) datasets
from confens.classifiers.Classifier import XGB, UnsupervisedClassifier
from confens.classifiers.ConfidenceBagging import ConfidenceBagging
from confens.classifiers.ConfidenceBoosting import ConfidenceBoosting
# The PYOD library contains implementations of unsupervised classifiers.
# Works only with anomaly detection (no multi-class)
# ------- GLOBAL VARS -----------
from confens.metrics.EnsembleMetric import SharedFaultMetric, DisagreementMetric

CSV_FOLDER = "sample_data"
# Name of the column that contains the label in the tabular (CSV) dataset
LABEL_NAME = 'multilabel'
# Name of the 'normal' class in datasets. This will be used only for binary classification (anomaly detection)
NORMAL_TAG = 'normal'
# Name of the file in which outputs of the analysis will be saved
SCORES_FILE = "cboost_wdiversity.csv"
# Percantage of test data wrt train data
TT_SPLIT = 0.5
# True if debug information needs to be shown
VERBOSE = True
# True if we want to conduct anomaly detection.
# This transforms multi-class labels into binary labels (rule: normal class vs others)
BINARIZE = True

# Set random seed for reproducibility
random.seed(42)
numpy.random.seed(42)


# --------- SUPPORT FUNCTIONS ---------------


def current_milli_time():
    """
    gets the current time in ms
    :return: a long int
    """
    return round(time.time() * 1000)


def get_learners(cont_perc):
    """
    Function to get a learner to use, given its string tag
    :param cont_perc: percentage of anomalies in the training set, required for unsupervised classifiers from PYOD
    :return: the list of classifiers to be trained
    """
    base_learners = [
        XGB(n_estimators=30),
        DecisionTreeClassifier(),
        Pipeline([("norm", MinMaxScaler()), ("gnb", GaussianNB())]),
        RandomForestClassifier(n_estimators=30),
        LinearDiscriminantAnalysis(),
        LogisticRegression(),
        ExtraTreesClassifier(n_estimators=30),
        LogitBoost(n_estimators=30)
    ]

    # If binary classification, we can use unsupervised classifiers also
    cont_alg = cont_perc if cont_perc < 0.5 else 0.5
    base_learners.extend([
        UnsupervisedClassifier(PCA(contamination=cont_alg)),
        UnsupervisedClassifier(INNE(contamination=cont_alg, n_estimators=10)),
        UnsupervisedClassifier(IForest(contamination=cont_alg, n_estimators=10)),
        UnsupervisedClassifier(HBOS(contamination=cont_alg, n_bins=30)),
        UnsupervisedClassifier(CBLOF(contamination=cont_alg, alpha=0.75, beta=3, n_jobs=-1)),
    ])

    learners = []
    for clf in base_learners:
        learners.append(clf)
        for n_base in [5, 10, 20]:
            for s_ratio in [0.2, 0.5, 0.7]:
                learners.append(ConfidenceBagging(clf=clf, n_base=n_base, sampling_ratio=s_ratio,
                                                  max_features=0.7, weighted=True))
                for n_decisors in [int(n_base / 2)]:
                    for mf in [0.5, 0.7]:
                        learners.append(ConfidenceBagging(clf=clf, n_base=n_base, n_decisors=n_decisors,
                                                          sampling_ratio=s_ratio, max_features=mf, weighted=True))
            for conf_thr in [0.8, 0.5]:
                for s_ratio in [0.3, 0.5]:
                    for w in [False, True]:
                        learners.append(ConfidenceBoosting(clf=clf, n_base=n_base, learning_rate=2,
                                                           sampling_ratio=s_ratio,
                                                           conf_thr=conf_thr, weighted=w))

    return learners


# ----------------------- MAIN ROUTINE ---------------------


if __name__ == '__main__':

    existing_exps = None
    if os.path.exists(SCORES_FILE):
        existing_exps = pandas.read_csv(SCORES_FILE)
        existing_exps = existing_exps.loc[:, ['dataset_tag', 'clf']]
    else:
        with open(SCORES_FILE, 'w') as f:
            f.write("dataset_tag,clf,binary,tt_split,acc,misc,mcc,ok_conf,misc_conf,time,model_size,disagreement,sharedfault\n")

    # Iterating over CSV files in folder
    for dataset_file in os.listdir(CSV_FOLDER):
        full_name = os.path.join(CSV_FOLDER, dataset_file)
        if full_name.endswith(".csv"):

            # if file is a CSV, it is assumed to be a dataset to be processed
            df = pandas.read_csv(full_name, sep=",")
            df = df.sample(frac=1.0)
            if len(df.index) > 80000:
                df = df.iloc[:80000, :]
            if VERBOSE:
                print("\n------------ DATASET INFO -----------------")
                print("Data Points in Dataset '%s': %d" % (dataset_file, len(df.index)))
                print("Features in Dataset: " + str(len(df.columns)))

            # Filling NaN and Handling (Removing) constant features
            df = df.fillna(0)
            df = df.loc[:, df.nunique() > 1]
            if VERBOSE:
                print("Features in Dataframe after removing constant ones: " + str(len(df.columns)))

            features_no_cat = df.select_dtypes(exclude=['object']).columns
            if VERBOSE:
                print("Features in Dataframe after non-numeric ones (including label): " + str(len(features_no_cat)))

            # Binarize if needed (for anomaly detection you need a 2-class problem, requires one of the classes to have NORMAL_TAG)
            normal_perc = None
            y = df[LABEL_NAME].to_numpy()
            if BINARIZE:
                y = numpy.where(df[LABEL_NAME] == "normal", 0, 1)
                if VERBOSE:
                    normal_frame = df.loc[df[LABEL_NAME] == NORMAL_TAG]
                    normal_perc = len(normal_frame.index) / len(df.index)
                    print("Normal data: " + str(len(normal_frame.index)) + " items (" +
                          "{:.3f}".format(100.0 * normal_perc) + "%)")
            elif VERBOSE:
                print("Dataset contains %d Classes" % len(numpy.unique(y)))

            # Set up train test split excluding categorical values that some algorithms cannot handle
            # 1-Hot-Encoding or other approaches may be used instead of removing
            x_no_cat = df.select_dtypes(exclude=['object']).to_numpy()
            x_train, x_test, y_train, y_test = ms.train_test_split(x_no_cat, y, test_size=TT_SPLIT, shuffle=True)

            if VERBOSE:
                print('-------------------- CLASSIFIERS -----------------------')

            # Loop for training and testing each learner specified by LEARNER_TAGS
            contamination = 1 - normal_perc if normal_perc is not None else None
            learners = get_learners(contamination)
            i = 1
            for classifier in learners:

                # Getting classifier Name
                clf_name = classifier.classifier_name() if hasattr(classifier,
                                                                   'classifier_name') else classifier.__class__.__name__
                if clf_name == 'Pipeline':
                    keys = list(classifier.named_steps.keys())
                    clf_name = str(keys) if len(keys) != 2 else str(keys[1]).upper()

                if existing_exps is not None and (((existing_exps['dataset_tag'] == full_name) &
                                                   (existing_exps['clf'] == clf_name)).any()):
                    print('%d/%d Skipping classifier %s, already in the results' % (i, len(learners), clf_name))

                else:
                    # Training the algorithm to get a model
                    start_time = current_milli_time()
                    classifier.fit(x_train, y_train)

                    # Quantifying size of the model
                    dump(classifier, "clf_dump.bin", compress=9)
                    size = os.stat("clf_dump.bin").st_size
                    os.remove("clf_dump.bin")

                    # Computing metrics
                    y_pred = classifier.predict(x_test)
                    if hasattr(classifier, 'predict_confidence') and callable(classifier.predict_confidence):
                        y_conf = classifier.predict_confidence(x_test)
                    else:
                        y_proba = classifier.predict_proba(x_test)
                        y_conf = numpy.max(y_proba, axis=1)
                    conf_ok = y_conf[numpy.where(y_pred == y_test)[0]]
                    conf_ok = [0.5] if len(conf_ok) == 0 else conf_ok
                    conf_ok_metrics = [numpy.min(conf_ok), numpy.median(conf_ok), numpy.average(conf_ok),
                                       numpy.max(conf_ok)]
                    conf_misc = y_conf[numpy.where(y_pred != y_test)[0]]
                    conf_misc = [0.5] if len(conf_misc) == 0 else conf_misc
                    conf_misc_metrics = [numpy.min(conf_misc), numpy.median(conf_misc), numpy.average(conf_misc),
                                         numpy.max(conf_misc)]

                    acc = metrics.accuracy_score(y_test, y_pred)
                    misc = int((1 - acc) * len(y_test))
                    mcc = abs(metrics.matthews_corrcoef(y_test, y_pred))

                    # Diversity
                    if hasattr(classifier, "get_diversity"):
                        diversity_dict = classifier.get_diversity(x_test, y_test, [DisagreementMetric(), SharedFaultMetric()])
                    else:
                        diversity_dict = {}

                    if BINARIZE:
                        # Prints metrics for binary classification + train time and model size
                        tn, fp, fn, tp = metrics.confusion_matrix(y_test, y_pred).ravel()
                        print('%d/%d %s\t-> TP: %d, TN: %d, FP: %d, FN: %d, Accuracy: %.3f, MCC: %.3f, Conf Diff: %.3f - train time: %d ms'
                              ' - model size: %.3f KB - shf: %.3f' %
                              (i, len(learners), clf_name, tp, tn, fp, fn, acc, mcc,
                               (conf_ok_metrics[2] - conf_misc_metrics[2]),
                               current_milli_time() - start_time, size / 1000.0,
                              (diversity_dict['SharedFault'] if 'SharedFault' in diversity_dict else 0.0)))
                    else:
                        # Prints just accuracy for multi-class classification problems, no confusion matrix
                        print('%d/%d %s\t-> Accuracy: %.3f, Conf Diff: %.3f - train time: %d ms - model size: %.3f KB'
                              % (i, len(learners), clf_name, acc, (conf_ok_metrics[2] - conf_misc_metrics[2]),
                                 current_milli_time() - start_time, size / 1000.0,
                              (diversity_dict['SharedFault'] if 'SharedFault' in diversity_dict else 0.0)))

                    # Updates CSV file form metrics of experiment
                    with open(SCORES_FILE, "a") as myfile:
                        # Prints result of experiment in CSV file
                        myfile.write(full_name + "," + clf_name + "," + str(BINARIZE) + "," +
                                     str(TT_SPLIT) + ',' + str(acc) + "," + str(misc) + "," + str(mcc) + "," +
                                     ";".join(["{:.4f}".format(met) for met in conf_ok_metrics]) + "," +
                                     ";".join(["{:.4f}".format(met) for met in conf_misc_metrics]) + "," +
                                     str(current_milli_time() - start_time) + "," + str(size) + "," +
                                     str((diversity_dict['Disagreement']/len(y_test) if 'Disagreement' in diversity_dict else 0.0)) + "," +
                                     str((diversity_dict['SharedFault']/len(y_test) if 'SharedFault' in diversity_dict else 0.0)) + "\n")

                classifier = None
                i += 1



import inspect
from collections.abc import Iterable

import numpy
from pyod.models.base import BaseDetector
from sklearn.base import is_classifier
from sklearn.utils.validation import check_is_fitted, check_array

def predict_proba(clf, X, get_base:bool = False):
    """
    Function to predict probabilities of a classifier
    Needed to overcome issues in pyod's predict_proba
    :param get_base: Tue if predictions of base-learners have to be returned as well
    :param clf: the classifier to be used
    :param X: the test set
    :return:
    """
    if isinstance(clf, BaseDetector):
        return predict_uns_proba(clf, X)
    elif 'get_base' in inspect.getfullargspec(clf.predict_proba)[0]:
        return clf.predict_proba(X, get_base=get_base)
    else:
        return clf.predict_proba(X)


def predict_uns_proba(uns_clf, X):
    """
    Method to compute probabilities of predicted classes.
    It has to e overridden since PYOD's implementation of predict_proba is wrong
    :return: array of probabilities for each classes
    """

    # Check if fit has been called
    check_is_fitted(uns_clf)
    X = check_array(X)
    probs = numpy.zeros((X.shape[0], 2))
    pred_score = uns_clf.decision_function(X)
    if numpy.isfinite(pred_score).all():
        if isinstance(uns_clf.contamination, (float, int)) and numpy.isfinite(uns_clf.threshold_):
            pred_thr = pred_score - uns_clf.threshold_
        else:
            pred_thr = pred_score
        min_pt = min(pred_thr)
        max_pt = max(pred_thr)
        anomaly = pred_thr > 0
        cont = numpy.asarray([pred_thr[i] / max_pt if anomaly[i] else (pred_thr[i] / min_pt if min_pt != 0 else 0.2)
                              for i in range(0, len(pred_thr))])
        probs[:, 0] = 0.5 + cont / 2
        probs[:, 1] = 1 - probs[:, 0]
        probs[anomaly, 0], probs[anomaly, 1] = probs[anomaly, 1], probs[anomaly, 0]
    else:
        probs[:, 0] = 0.999
        probs[:, 1] = 0.001
    return probs


def get_classifier_name(clf_object):
    """
    Gets a string representing the classifier name
    :param clf_object: the object meant to be a classifier
    :return: a string
    """
    clf_name = ""
    if clf_object is not None:
        if is_classifier(clf_object) or isinstance(clf_object, BaseDetector):
            clf_name = get_single_classifier_name(clf_object)
        elif isinstance(clf_object, Iterable):
            if len(clf_object) < 5:
                for clf_item in clf_object:
                    clf_name = clf_name + (get_single_classifier_name(clf_item) if is_classifier(clf_item) else "?") + "@"
                clf_name = clf_name[0:-1]
            else:
                clf_name = clf_name + str(len(clf_object)) + "_algs"
        else:
            clf_name = str(clf_object)
        if hasattr(clf_object, "base_estimator") and hasattr(clf_object, "n_estimators"):
            clf_name = clf_name + "(" + get_single_classifier_name(clf_object.base_estimator) + ";" \
                       + str(clf_object.n_estimators) + ")"
        if hasattr(clf_object, "estimators"):
            if len(clf_object.estimators) < 5:
                clf_name = clf_name + "(" + "@".join([get_single_classifier_name(clf) for clf in clf_object.estimators]) + ";" \
                       + str(len(clf_object.estimators)) + ")"
            else:
                clf_name = clf_name + "(" + str(len(clf_object.estimators)) + ")"
    return clf_name


def get_single_classifier_name(clf_object):
    """
    Gets a string representing the classifier name, assuming the object contains a single classifier
    :param clf_object: the object meant to be a classifier
    :return: a string
    """
    if hasattr(clf_object, "classifier_name") and callable(clf_object.classifier_name):
        clf_name = clf_object.classifier_name()
        if clf_name == 'Pipeline':
            for x in list(clf_object.named_steps.keys()):
                if is_classifier(clf_object[x]):
                    clf_name = get_single_classifier_name(clf_object[x])
    elif isinstance(clf_object, tuple):
        clf_name = str(clf_object[0])
        for x in clf_object:
            if is_classifier(x):
                clf_name = get_single_classifier_name(x)
    else:
        clf_name = clf_object.__class__.__name__
        if clf_name == 'Pipeline':
            for x in list(clf_object.named_steps.keys()):
                if is_classifier(clf_object[x]):
                    clf_name = get_single_classifier_name(clf_object[x])
    return clf_name


def predict_confidence(clf, X):
    """
    Method to compute the confidence in predictions of a classifier
    :param clf: the classifier
    :param X: the test set
    :return: array of confidence scores
    """
    c_conf = None
    if isinstance(clf, BaseDetector):
        y_proba = predict_proba(clf, X)
        c_conf = numpy.max(y_proba, axis=1)
    if is_classifier(clf):
        if hasattr(clf, 'predict_confidence') and callable(clf.predict_confidence):
            c_conf = clf.predict_confidence(X)
        else:
            y_proba = predict_proba(clf, X)
            c_conf = numpy.max(y_proba, axis=1)
    return c_conf



import numpy

from confens.metrics.DiversityMetric import QStatDiversity, SigmaDiversity, Disagreement


def get_default():
    return [QStatMetric(), SigmaMetric(), CoupleDisagreementMetric(),
            DisagreementMetric(), SharedFaultMetric()]


class EnsembleMetric:
    """
   Abstract Class for ensemble metrics.
   """

    def get_name(self):
        """
        Returns the name of the metric (as string)
        """
        pass

    def compute_diversity(self, clf_predictions, test_y):
        """
        Abstract method to compute diversity metric
        :param clf_predictions: predictions of clfs in the ensemble
        :param test_y: reference test labels
        :return: a float value
        """
        pass


class QStatMetric(EnsembleMetric):
    """

    """

    def __init__(self):
        """
        Constructor Method
        """
        return

    def compute_diversity(self, clf_predictions, test_y):
        """
        Abstract method to compute diversity metric
        :param clf_predictions: predictions of clfs in the ensemble
        :param test_y: reference test labels
        :return: a float value
        """
        n_clfs = clf_predictions.shape[1]
        qd = QStatDiversity()
        diversities = [qd.compute_diversity(clf_predictions[:, i], clf_predictions[:, j], test_y)
                       for i in range(0, n_clfs - 1) for j in range(i + 1, n_clfs)]
        return 2 * sum(diversities) / (n_clfs * (n_clfs - 1))

    def get_name(self):
        return 'QStat'


class SigmaMetric(EnsembleMetric):
    """

    """

    def __init__(self):
        """
        Constructor Method
        """
        return

    def compute_diversity(self, clf_predictions, test_y):
        """
        Abstract method to compute diversity metric
        :param clf_predictions: predictions of clfs in the ensemble
        :param test_y: reference test labels
        :return: a float value
        """
        n_clfs = clf_predictions.shape[1]
        qd = SigmaDiversity()
        diversities = [qd.compute_diversity(clf_predictions[:, i], clf_predictions[:, j], test_y)
                       for i in range(0, n_clfs - 1) for j in range(i + 1, n_clfs)]
        return 2 * sum(diversities) / (n_clfs * (n_clfs - 1))

    def get_name(self):
        return 'Sigma'


class CoupleDisagreementMetric(EnsembleMetric):
    """

    """

    def __init__(self):
        """
        Constructor Method
        """
        return

    def compute_diversity(self, clf_predictions, test_y):
        """
        Abstract method to compute diversity metric
        :param clf_predictions: predictions of clfs in the ensemble
        :param test_y: reference test labels
        :return: a float value
        """
        n_clfs = clf_predictions.shape[1]
        qd = Disagreement()
        diversities = [qd.compute_diversity(clf_predictions[:, i], clf_predictions[:, j], test_y)
                       for i in range(0, n_clfs - 1) for j in range(i + 1, n_clfs)]
        return numpy.mean(diversities)

    def get_name(self):
        return 'CoupleDisagreement'


class DisagreementMetric(EnsembleMetric):
    """

    """

    def __init__(self):
        """
        Constructor Method
        """
        return

    def compute_diversity(self, clf_predictions, test_y):
        """
        Abstract method to compute diversity metric
        :param clf_predictions: predictions of clfs in the ensemble
        :param test_y: reference test labels
        :return: a float value
        """
        clf_hits = numpy.asarray(
            [(clf_predictions[:, i] == test_y) for i in range(0, clf_predictions.shape[1])]).transpose()
        agreements = [numpy.all(clf_hits[i, :] == clf_hits[i, 0]) for i in range(0, clf_predictions.shape[0])]
        return len(test_y) - sum(agreements)

    def get_name(self):
        return 'Disagreement'


class SharedFaultMetric(EnsembleMetric):
    """

    """

    def __init__(self):
        """
        Constructor Method
        """
        return

    def compute_diversity(self, clf_predictions, test_y):
        """
        Abstract method to compute diversity metric
        :param clf_predictions: predictions of clfs in the ensemble
        :param test_y: reference test labels
        :return: a float value
        """
        comp_array = [False for i in range(0, clf_predictions.shape[1])]
        clf_hits = numpy.asarray(
            [(clf_predictions[:, i] == test_y) for i in range(0, clf_predictions.shape[1])]).transpose()
        shared_faults = numpy.asarray(
            [numpy.all(clf_hits[i, :] == comp_array) for i in range(0, clf_predictions.shape[0])])
        return sum(shared_faults)

    def get_name(self):
        return 'SharedFault'



import copy

import numpy

from confens.classifiers.ConfidenceEnsemble import ConfidenceEnsemble
from confens.utils.classifier_utils import get_classifier_name, predict_confidence


class ConfidenceBoosting(ConfidenceEnsemble):
    """
    Class for creating Confidence Boosting ensembles
    """

    def __init__(self, clf, n_base: int = 20, learning_rate: float = 3.0,
                 sampling_ratio: float = 0.7, relative_boost_thr: float = 0.8, static_boost_thr: float = None,
                 conf_thr: float = None, perc_decisors: float = 1.0,
                 n_decisors: int = None, weighted: bool = True):
        """
        Constructor
        :param clf: the algorithm to be used for creating base learners
        :param n_base: number of base learners (= size of the ensemble)
        :param learning_rate: learning rate for updating dataset weights
        :param sampling_ratio: percentage of the dataset to be used at each iteration
        :param boost_thr: threshold of acceptance for confidence scores. It is the percentile of confidence scores of the base estimator that are considered "confident enough"
        :param static_boost_thr: static threshold of acceptance for confidence scores. Lower confidence means untrustable result
        :param conf_thr: float value for confidence threshold
        :param perc_decisors: percentage of base learners to be used for prediction
        :param n_decisors: number of base learners to be used for prediction
        :param weighted: True if prediction has to be computed as a weighted sum of probabilities
        """
        super().__init__(clf, n_base, conf_thr, perc_decisors, n_decisors, weighted)

        # Boosting thresholds
        self.relative_boost_thr = relative_boost_thr
        self.static_boost_thr = static_boost_thr

        # Other ConfBoost parameters
        if learning_rate is not None:
            self.learning_rate = learning_rate
        else:
            self.learning_rate = 2
        if sampling_ratio is not None:
            self.sampling_ratio = sampling_ratio
        else:
            self.sampling_ratio = 1 / n_base ** (1 / 2)

    def fit_ensemble(self, X, y=None):
        """
        Training function for the confidence boosting ensemble
        :param y: labels of the train set (optional, not required for unsupervised learning)
        :param X: train set
        """
        train_n = len(X)
        samples_n = int(train_n * self.sampling_ratio)
        weights = numpy.full(train_n, 1 / train_n)

        for learner_index in range(0, self.n_base):

            # Draw samples
            sample_x, sample_y = self.draw_samples(X, y, samples_n, weights)

            # Train learner
            learner = copy.deepcopy(self.clf_list[learner_index % len(self.clf_list)])
            learner.fit(sample_x, sample_y)
            if hasattr(learner, "X_"):
                learner.X_ = None
            if hasattr(learner, "y_"):
                learner.y_ = None
            self.estimators_.append(learner)

            # Derive boost threshold
            y_conf = predict_confidence(learner, X)
            if self.static_boost_thr is not None and 0 < self.static_boost_thr < 1:
                p_thr = self.static_boost_thr
            else:
                p_thr = numpy.sort(y_conf)[int((1-self.relative_boost_thr)*len(y_conf))] if y_conf is not None else 0.8

            # Update Weights
            update_flag = numpy.where(y_conf >= p_thr, 0, 1)
            weights = weights * (1 + self.learning_rate * update_flag)
            weights = weights / sum(weights)

    def classifier_name(self):
        """
        Gets classifier name as string
        :return: the classifier name
        """
        clf_name = get_classifier_name(self.clf)
        if self.weighted:
            return "ConfidenceBoosterWeighted(" + str(clf_name) + "-" + \
                   str(self.n_base) + "-" + str(self.relative_boost_thr) + "-" + str(self.static_boost_thr) + "-" + \
                   str(self.learning_rate) + "-" + str(self.sampling_ratio) + "-" + \
                   str(self.conf_thr) + "-" + str(self.perc_decisors) + "-" + str(self.n_decisors) + ")"
        else:
            return "ConfidenceBooster(" + str(clf_name) + "-" + \
                   str(self.n_base) + "-" + str(self.relative_boost_thr) + "-" + str(self.static_boost_thr) + "-" + \
                   str(self.learning_rate) + "-" + str(self.sampling_ratio) + "-" + \
                   str(self.conf_thr) + "-" + str(self.perc_decisors) + "-" + str(self.n_decisors) + ")"



# Support libs
import random

import numpy
import pandas
import sklearn.metrics as metrics
import sklearn.model_selection as ms
# Used to save a classifier and measure its size in KB
from pyod.models.copod import COPOD
from pyod.models.feature_bagging import FeatureBagging
from pyod.models.hbos import HBOS
from pyod.models.pca import PCA
# Name of the folder in which look for tabular (CSV) datasets
from pyod.models.suod import SUOD

from confens.classifiers.ConfidenceBagging import ConfidenceBagging
from confens.classifiers.ConfidenceBoosting import ConfidenceBoosting
# Scikit-Learn algorithms
# The PYOD library contains implementations of unsupervised classifiers.
# Works only with anomaly detection (no multi-class)
# ------- GLOBAL VARS -----------
from confens.utils.general_utils import get_classifier_name

CSV_FILE = "sample_data/sample_data_arancino.csv"
# Name of the column that contains the label in the tabular (CSV) dataset
LABEL_NAME = 'multilabel'
# Name of the 'normal' class in datasets. This will be used only for binary classification (anomaly detection)
NORMAL_TAG = 'normal'
# Percantage of test data wrt train data
TT_SPLIT = 0.5
# True if debug information needs to be shown
VERBOSE = True

# Set random seed for reproducibility
random.seed(42)
numpy.random.seed(42)

# ----------------------- MAIN ROUTINE ---------------------
# This files shows an example of confidence ensembles for unsupervised learning.
# Can be used as is, just change the details to load the dataset

if __name__ == '__main__':

    # if file is a CSV, it is assumed to be a dataset to be processed
    df = pandas.read_csv(CSV_FILE, sep=",", nrows=10000)
    if VERBOSE:
        print("\n------------ DATASET INFO -----------------")
        print("Data Points in Dataset '%s': %d" % (CSV_FILE, len(df.index)))
        print("Features in Dataset: " + str(len(df.columns)))

    # Set up train test split excluding categorical values that some algorithms cannot handle
    # 1-Hot-Encoding or other approaches may be used instead of removing
    y = numpy.where(df[LABEL_NAME] == "normal", 0, 1)
    an_perc = numpy.average(y)
    x_train, x_test, y_train, y_test = ms.train_test_split(df.drop(columns=[LABEL_NAME]).to_numpy(), y,
                                                           test_size=TT_SPLIT, shuffle=True)
    # Creating classifiers
    # Note that unsupervised classifiers in PYOD require a contamination
    classifiers = [PCA(contamination=an_perc),
                   ConfidenceBagging(clf=PCA(contamination=an_perc)),
                   ConfidenceBoosting(clf=PCA(contamination=an_perc)),
                   FeatureBagging(contamination=an_perc, base_estimator=ConfidenceBagging(clf=HBOS(contamination=an_perc)), n_estimators=5),
                   FeatureBagging(contamination=an_perc, base_estimator=ConfidenceBoosting(clf=HBOS(contamination=an_perc)), n_estimators=5),
                   SUOD(contamination=an_perc, base_estimators=[PCA(contamination=an_perc),
                                                                ConfidenceBagging(clf=COPOD(contamination=an_perc)),
                                                                ConfidenceBoosting(clf=COPOD(contamination=an_perc))])]
    for classifier in classifiers:
        #cb_clf = ConfidenceBoosting(clf=classifier, n_base=10, learning_rate=2,
        #                            sampling_ratio=0.5, conf_thr=0.8)

        # Exercising Classifier
        clf_name = get_classifier_name(classifier)
        classifier.fit(x_train)
        clf_pred = classifier.predict(x_test)

        # Exercising ConfBoost
        #cb_clf.fit(x_train)
        #cb_pred = cb_clf.predict(x_test)

        print('%s has accuracy of %.3f' %
              (clf_name, metrics.accuracy_score(y_test, clf_pred)))#, metrics.accuracy_score(y_test, cb_pred)))



# Support libs
import os
import random
import time

import numpy
import pandas
import sklearn.metrics as metrics
import sklearn.model_selection as ms
# Used to save a classifier and measure its size in KB
from joblib import dump
from logitboost import LogitBoost
from pyod.models.cblof import CBLOF
from pyod.models.hbos import HBOS
from pyod.models.iforest import IForest
from pyod.models.inne import INNE
from pyod.models.pca import PCA
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
# Scikit-Learn algorithms
from sklearn.ensemble import RandomForestClassifier, ExtraTreesClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler
from sklearn.tree import DecisionTreeClassifier

# Name of the folder in which look for tabular (CSV) datasets
from confens.classifiers.Classifier import XGB, UnsupervisedClassifier
from confens.classifiers.ConfidenceBagging import ConfidenceBagging
from confens.classifiers.ConfidenceBoosting import ConfidenceBoosting
# The PYOD library contains implementations of unsupervised classifiers.
# Works only with anomaly detection (no multi-class)
# ------- GLOBAL VARS -----------
from confens.metrics.EnsembleMetric import DisagreementMetric, SharedFaultMetric

CSV_FOLDER = "sample_data"
# Name of the column that contains the label in the tabular (CSV) dataset
LABEL_NAME = 'multilabel'
# Name of the 'normal' class in datasets. This will be used only for binary classification (anomaly detection)
NORMAL_TAG = 'normal'
# Name of the file in which outputs of the analysis will be saved
SCORES_FILE = "test_scores_unknowns.csv"
# Percantage of test data wrt train data
TT_SPLIT = 0.5
# True if debug information needs to be shown
VERBOSE = True

# Set random seed for reproducibility
random.seed(42)
numpy.random.seed(42)


# --------- SUPPORT FUNCTIONS ---------------


def current_milli_time():
    """
    gets the current time in ms
    :return: a long int
    """
    return round(time.time() * 1000)


def get_learners(cont_perc):
    """
    Function to get a learner to use, given its string tag
    :param cont_perc: percentage of anomalies in the training set, required for unsupervised classifiers from PYOD
    :return: the list of classifiers to be trained
    """
    base_learners = [
        XGB(n_estimators=30),
        DecisionTreeClassifier(),
        Pipeline([("norm", MinMaxScaler()), ("gnb", GaussianNB())]),
        RandomForestClassifier(n_estimators=30),
        LinearDiscriminantAnalysis(),
        LogisticRegression(),
        ExtraTreesClassifier(n_estimators=30),
        LogitBoost(n_estimators=30)
    ]

    # If binary classification, we can use unsupervised classifiers also
    cont_alg = cont_perc if cont_perc < 0.5 else 0.5
    base_learners.extend([
        UnsupervisedClassifier(PCA(contamination=cont_alg)),
        UnsupervisedClassifier(INNE(contamination=cont_alg, n_estimators=10)),
        UnsupervisedClassifier(IForest(contamination=cont_alg, n_estimators=10)),
        UnsupervisedClassifier(HBOS(contamination=cont_alg, n_bins=30)),
        UnsupervisedClassifier(CBLOF(contamination=cont_alg, alpha=0.75, beta=3, n_jobs=-1)),
    ])

    learners = []
    for clf in base_learners:
        learners.append(clf)
        for n_base in [10]:
            for s_ratio in [0.2, 0.5]:
                learners.append(ConfidenceBagging(clf=clf, n_base=n_base, sampling_ratio=s_ratio,
                                                  max_features=0.7, weighted=True))
                for n_decisors in [int(n_base / 2)]:
                    learners.append(ConfidenceBagging(clf=clf, n_base=n_base, n_decisors=n_decisors,
                                                      sampling_ratio=s_ratio, max_features=0.7))
            for conf_thr in [0.9, 0.8]:
                for s_ratio in [0.1, 0.3, 0.5]:
                    for w in [False, True]:
                        learners.append(ConfidenceBoosting(clf=clf, n_base=n_base, learning_rate=2,
                                                           sampling_ratio=s_ratio,
                                                           conf_thr=conf_thr, weighted=w))

    return learners


# ----------------------- MAIN ROUTINE ---------------------
# This script replicates experiments done for testing the robustness of confidence ensembles

if __name__ == '__main__':

    existing_exps = None
    if os.path.exists(SCORES_FILE):
        existing_exps = pandas.read_csv(SCORES_FILE)
        existing_exps = existing_exps.loc[:, ['dataset_tag', 'unknown', 'clf']]
    else:
        with open(SCORES_FILE, 'w') as f:
            f.write("dataset_tag,unknown,clf,len_test,len_unk,acc,mcc,rec_unk,ok_confs,misc_confs,unk_ok_confs,"
                    "unk_misc_confs,time,model_size,disagreement,sharedfault\n")

            # Iterating over CSV files in folder
    for dataset_file in os.listdir(CSV_FOLDER):
        full_name = os.path.join(CSV_FOLDER, dataset_file)
        if full_name.endswith(".csv"):

            # if file is a CSV, it is assumed to be a dataset to be processed
            df = pandas.read_csv(full_name, sep=",")
            df = df.sample(frac=1.0)
            if len(df.index) > 100000:
                df = df.iloc[:100000, :]
            if VERBOSE:
                print("\n------------ DATASET INFO -----------------")
                print("Data Points in Dataset '%s': %d" % (dataset_file, len(df.index)))
                print("Features in Dataset: " + str(len(df.columns)))

            # Filling NaN and Handling (Removing) constant features
            df = df.fillna(0)
            df = df.loc[:, df.nunique() > 1]
            if VERBOSE:
                print("Features in Dataframe after removing constant ones: " + str(len(df.columns)))

            features_no_cat = df.select_dtypes(exclude=['object']).columns
            if VERBOSE:
                print("Features in Dataframe after non-numeric ones (including label): " + str(len(features_no_cat)))

            # Check if dataset has more than 2 classes
            y = df[LABEL_NAME].to_numpy()
            classes = numpy.unique(y)
            if len(classes) > 2:

                print("Dataset contains %d Classes" % len(numpy.unique(y)))

                # Set up train test split excluding categorical values that some algorithms cannot handle
                # 1-Hot-Encoding or other approaches may be used instead of removing
                x_no_cat = df.select_dtypes(exclude=['object']).to_numpy()
                x_tr, x_test, y_tr, y_te = ms.train_test_split(x_no_cat, y, test_size=TT_SPLIT, shuffle=True)

                index = 0
                tot = (len(classes) - 1) * len(get_learners(0.5))
                # Iterate over anomalies
                for anomaly in classes:

                    # Check if class is an anomaly
                    if anomaly != NORMAL_TAG:

                        print("\n--------- ANALYSIS WITH '%s' AS UNKNOWN -------------" % anomaly)
                        train_indexes_to_remove = numpy.asarray(numpy.where(y_tr == anomaly)[0])
                        x_train = numpy.delete(x_tr, train_indexes_to_remove, axis=0)
                        y_train = numpy.delete(y_tr, train_indexes_to_remove, axis=0)
                        y_test = y_te
                        test_indexes_anomaly = numpy.asarray(numpy.where(y_te == anomaly)[0])
                        x_test_unknowns = x_test[test_indexes_anomaly, :]
                        y_test_unknowns = [1 for _ in range(0, len(test_indexes_anomaly))]

                        # Binarize (for anomaly detection you need a 2-class problem,
                        # requires one of the classes to have NORMAL_TAG)

                        y_train = numpy.where(y_train == NORMAL_TAG, 0, 1)
                        y_test = numpy.where(y_te == NORMAL_TAG, 0, 1)
                        normal_frame = df.loc[df[LABEL_NAME] == NORMAL_TAG]
                        normal_perc = len(normal_frame.index) / len(df.index)

                        if VERBOSE:
                            print('-------------------- CLASSIFIERS -----------------------')

                        # Loop for training and testing each learner specified by LEARNER_TAGS
                        contamination = 1 - normal_perc if normal_perc is not None else None
                        learners = get_learners(contamination)
                        i = 1
                        for classifier in learners:

                            # Getting classifier Name
                            clf_name = classifier.classifier_name() if hasattr(classifier,
                                                                               'classifier_name') else classifier.__class__.__name__
                            if clf_name == 'Pipeline':
                                keys = list(classifier.named_steps.keys())
                                clf_name = str(keys) if len(keys) != 2 else str(keys[1]).upper()

                            if existing_exps is not None and (((existing_exps['dataset_tag'] == full_name) &
                                                               (existing_exps['unknown'] == anomaly) &
                                                               (existing_exps['clf'] == clf_name)).any()):
                                print('%d/%d Skipping classifier %s, already in the results' % (
                                    i, len(learners), clf_name))

                            else:

                                # Training the algorithm to get a model
                                start_time = current_milli_time()
                                classifier.fit(x_train, y_train)

                                # Quantifying size of the model
                                dump(classifier, "clf_dump.bin", compress=9)
                                size = os.stat("clf_dump.bin").st_size
                                os.remove("clf_dump.bin")

                                # Computing metrics
                                y_pred = classifier.predict(x_test)
                                if hasattr(classifier, 'predict_confidence') and callable(
                                        classifier.predict_confidence):
                                    y_conf = classifier.predict_confidence(x_test)
                                else:
                                    y_proba = classifier.predict_proba(x_test)
                                    y_conf = numpy.max(y_proba, axis=1)
                                conf_ok = y_conf[numpy.where(y_pred == y_test)[0]]
                                conf_ok = [0.5] if len(conf_ok) == 0 else conf_ok
                                conf_ok_metrics = [numpy.min(conf_ok), numpy.median(conf_ok), numpy.average(conf_ok),
                                                   numpy.max(conf_ok)]
                                conf_misc = y_conf[numpy.where(y_pred != y_test)[0]]
                                conf_misc = [0.5] if len(conf_misc) == 0 else conf_misc
                                conf_misc_metrics = [numpy.min(conf_misc), numpy.median(conf_misc),
                                                     numpy.average(conf_misc),
                                                     numpy.max(conf_misc)]
                                acc = metrics.accuracy_score(y_test, y_pred)
                                mcc = abs(metrics.matthews_corrcoef(y_test, y_pred))

                                # Computing metrics for unknowns
                                y_pred_unk = classifier.predict(x_test_unknowns)
                                rec_unk = numpy.average(y_test_unknowns == y_pred_unk)
                                if hasattr(classifier, 'predict_confidence') and callable(
                                        classifier.predict_confidence):
                                    y_conf = classifier.predict_confidence(x_test_unknowns)
                                else:
                                    y_proba = classifier.predict_proba(x_test_unknowns)
                                    y_conf = numpy.max(y_proba, axis=1)
                                conf_ok = y_conf[numpy.where(y_pred_unk == y_test_unknowns)[0]]
                                conf_ok = [0.5] if len(conf_ok) == 0 else conf_ok
                                confunk_ok_metrics = [numpy.min(conf_ok), numpy.median(conf_ok), numpy.average(conf_ok),
                                                      numpy.max(conf_ok)]
                                conf_misc = y_conf[numpy.where(y_pred_unk != y_test_unknowns)[0]]
                                conf_misc = [0.5] if len(conf_misc) == 0 else conf_misc
                                confunk_misc_metrics = [numpy.min(conf_misc), numpy.median(conf_misc),
                                                        numpy.average(conf_misc),
                                                        numpy.max(conf_misc)]

                                # Diversity
                                if hasattr(classifier, "get_diversity"):
                                    diversity_dict = classifier.get_diversity(x_test, y_test, [DisagreementMetric(),
                                                                                               SharedFaultMetric()])
                                else:
                                    diversity_dict = {}

                                # Prints metrics for binary classification + train time and model size
                                print(
                                    '%d/%d %s\t-> ACC: %.3f, MCC: %.3f, REC_UNK: %.3f, Conf Diff: %.3f, ConfUnk Diff: %.3f '
                                    '- train time: %d ms - model size: %.3f KB' %
                                    (index, tot, clf_name, acc, mcc, rec_unk,
                                     (conf_ok_metrics[2] - conf_misc_metrics[2]),
                                     (confunk_ok_metrics[2] - confunk_misc_metrics[2]),
                                     current_milli_time() - start_time, size / 1000.0))

                                # Updates CSV file form metrics of experiment
                                with open(SCORES_FILE, "a") as myfile:
                                    # Prints result of experiment in CSV file
                                    myfile.write(full_name + "," + str(anomaly) + "," + clf_name +
                                                 "," + str(len(y_test)) + ',' + str(len(y_test_unknowns)) + ',' +
                                                 str(acc) + "," + str(mcc) + "," + str(rec_unk) + "," +
                                                 ";".join(["{:.4f}".format(met) for met in conf_ok_metrics]) + "," +
                                                 ";".join(["{:.4f}".format(met) for met in conf_misc_metrics]) + "," +
                                                 ";".join(["{:.4f}".format(met) for met in confunk_ok_metrics]) + "," +
                                                 ";".join(
                                                     ["{:.4f}".format(met) for met in confunk_misc_metrics]) + "," +
                                                 str(current_milli_time() - start_time) + "," + str(size) + "," +
                                                 str((diversity_dict['Disagreement'] / len(
                                                     y_test) if 'Disagreement' in diversity_dict else 0.0)) + "," +
                                                 str((diversity_dict['SharedFault'] / len(
                                                     y_test) if 'SharedFault' in diversity_dict else 0.0)) + "\n")

                            classifier = None
                            index += 1
            else:
                print('Dataset does not have more than 2 classes, no way to simulating unknowns')



import copy
from collections.abc import Iterable
from multiprocessing.pool import ThreadPool

import numpy
from pyod.models.base import BaseDetector
from sklearn.base import is_classifier, BaseEstimator, ClassifierMixin
from sklearn.ensemble import RandomForestClassifier
from sklearn.utils import check_X_y, check_array
from sklearn.utils.multiclass import unique_labels
from sklearn.utils.validation import check_is_fitted

from confens.metrics.EnsembleMetric import get_default
from confens.utils.classifier_utils import predict_proba, predict_confidence, get_classifier_name


def define_bin_proba_thr(probs, cont: float = None) -> float:
    """
    Method for finding a confidence threshold based on the expected contamination (iterative)
    :param probs: probabilities to find threshold of
    :param target: the quantity to be used as reference for gettng to the threshold
    :return: a float value to be used as threshold to decide on anomalies (binary classification)
    """
    an_probas = numpy.sort(probs[:, 1])
    cutoff_index = int(len(an_probas)*(1-cont))
    thr = an_probas[cutoff_index]
    return thr


def define_proba_thr(probs, target: float = None, delta: float = 0.01) -> float:
    """
    Method for finding a confidence threshold based on the expected contamination (iterative)
    :param probs: probabilities to find threshold of
    :param target: the quantity to be used as reference for gettng to the threshold
    :param delta: the tolerance to stop recursion
    :return: a float value to be used as threshold for updating weights in boosting
    """
    target_cont = target
    p_thr = 0.5
    left_bound = 0.5
    right_bound = 1
    actual_cont = numpy.average(probs[:, 0] < p_thr)
    while abs(actual_cont - target_cont) > delta and abs(right_bound - left_bound) > 0.01:
        if actual_cont < target_cont:
            left_bound = p_thr
            p_thr = (p_thr + right_bound) / 2
        else:
            right_bound = p_thr
            p_thr = (p_thr + left_bound) / 2
        actual_cont = numpy.average(probs[:, 0] < p_thr)
    return p_thr


class ConfidenceEnsemble(ClassifierMixin, BaseEstimator):
    """
    Class for creating confidence ensembles
    """

    def __init__(self, clf, n_base: int = 10, conf_thr: float = None, perc_decisors: float = None,
                 n_decisors: int = None, weighted: bool = False):
        """
        Constructor
        :param clf: the algorithm(s) to be used for creating base learners
        :param n_base: number of base learners (= size of the ensemble)
        :param conf_thr: float value for confidence threshold
        :param perc_decisors: percentage of base learners to be used for prediction
        :param n_decisors: number of base learners to be used for prediction
        :param weighted: True if prediction has to be computed as a weighted sum of probabilities
        """
        super().__init__()
        self.clf = clf
        self._estimator_type = "classifier"
        self.feature_importances_ = None
        self.X_ = None
        self.y_ = None
        self.clf_list = []
        self.input_val = False
        self.weighted = weighted
        self.bin_proba_thr = None
        self.conf_thr = conf_thr
        self.contamination = clf.contamination if hasattr(clf, 'contamination') else None
        self.perc_decisors = perc_decisors
        self.n_decisors = n_decisors
        self.n_base = n_base
        self.estimators_ = []

    def validate_input(self):
        # Setting up classifiers to create base-learners
        self.clf = copy.deepcopy(self.clf) if self.clf is not None else None
        if is_classifier(self.clf) or isinstance(self.clf, BaseDetector):
            self.clf_list = [self.clf]
        elif isinstance(self.clf, Iterable):
            self.clf_list = []
            for clf_item in self.clf:
                if is_classifier(clf_item) or isinstance(clf_item, BaseDetector):
                    self.clf_list.append(clf_item)
                else:
                    print("Cant recognize object s a classifier")
            if len(self.clf_list) == 0:
                self.clf_list = [RandomForestClassifier(n_estimators=10)]
                print("clf is not a classifier. Using a 10-tree Random Forest as Base estimator")
        else:
            self.clf_list = [RandomForestClassifier(n_estimators=10)]
            print("clf is not a classifier. Using a 10-tree Random Forest as Base estimator")

        # N base estimators
        if self.n_base is None or self.n_base <= 1:
            print("Ensembles have to be at least 2")
            self.n_base = 10

        # Decisors
        if self.perc_decisors is not None and 0 < self.perc_decisors <= 1:
            if self.n_decisors is not None and 0 < self.n_decisors <= self.n_base:
                print('Both perc_decisors and n_decisors are specified, prioritizing perc_decisors')
            self.n_decisors = int(self.n_base * self.perc_decisors) if int(self.n_base * self.perc_decisors) > 0 else 1
        elif self.n_decisors is not None and 0 < self.n_decisors <= self.n_base:
            self.n_decisors = self.n_decisors
        elif self.conf_thr is None:
            self.n_decisors = self.n_base
        else:
            self.n_decisors = None

        # END
        self.input_val = True

    def fit(self, X, y=None):
        """
        Training function for the classifier
        :param y: labels of the train set (optional, not required for unsupervised learning)
        :param X: train set
        """
        if y is not None:
            X, y = check_X_y(X, y)
        else:
            X = check_array(X)
        self.classes_ = unique_labels(y) if y is not None else numpy.array([0, 1])
        if not self.input_val:
            self.validate_input()
        self.fit_ensemble(X, y)
        self.bin_proba_thr = define_bin_proba_thr(predict_proba(self, X), self.contamination) \
            if self.contamination is not None else 0.5

        # Compliance with SKLEARN and PYOD
        self.X_ = X[[0, 1], :]
        self.y_ = y
        self.decision_scores_ = self.decision_function(X)
        self.feature_importances_ = self.compute_feature_importances()

    def fit_ensemble(self, X, y=None):
        """
        Training function for the confidence ensemble: TO BE OVERRIDDEN
        :param y: labels of the train set (optional, not required for unsupervised learning)
        :param X: train set
        """
        pass

    def predict_base(self, X, learner_index):
        if hasattr(self, "feature_sets"):
            # ConfBag, each estimator uses a subset of features
            predictions = predict_proba(self.estimators_[learner_index], X[:, self.feature_sets[learner_index]])
        else:
            # ConfBoost, all estimators use all features
            predictions = predict_proba(self.estimators_[learner_index], X)
        return predictions

    def predict_proba(self, X, get_base:bool = False):
        """
        Function to assign probabilities to each class given a test set
        :param get_base: True if probabilities of base-learners have to be returned as well
        :param X: the test set
        :return: a numpy matrix
        """
        # Scoring probabilities and confidence
        proba_array = []
        conf_array = []
        base_dict = {}
        with ThreadPool(self.n_base) as p:
            proba_array = p.starmap(self.predict_base, [(X, i) for i in range(0, self.n_base)])
        for i in range(0, self.n_base):
            predictions = proba_array[i]
            base_dict[str(i) + '#' + get_classifier_name(self.estimators_[i])] = predictions
            conf_array.append(numpy.max(predictions, axis=1))
        # 3d matrix (clf, row, probability for class)
        proba_array = numpy.asarray(proba_array)
        # 2dim matrix (clf, confidence for row)
        conf_array = numpy.asarray(conf_array)

        # Compute final probabilities
        # Adjust probabilities with confidence if weighted
        if self.weighted:
            for i in range(0, X.shape[0]):
                proba_array[:, i, :] = (proba_array[:, i, :].T * conf_array[:, i]).T

        if self.n_decisors is not None:
            # Option 1: either n_decisors or perc_decisors is set, or conf_thr is None
            if self.n_decisors > self.n_base or self.n_decisors <= 0:
                # Checks for strange n_decisors values
                self.n_decisors = self.n_base
            if self.n_decisors == self.n_base:
                # Easy case, all contribute, no need to additional computation
                proba = numpy.average(proba_array, axis=0)
                proba = proba / numpy.sum(proba, axis=1).reshape(-1, 1)
            else:
                # Here it requires sorting confidences for understanding the "best" base-learners
                conf_array = conf_array.transpose()
                all_conf = -numpy.sort(-conf_array, axis=1)
                conf_thrs = all_conf[:, self.n_decisors - 1]
                mask = numpy.asarray(conf_array >= conf_thrs.reshape(-1, 1)).T
                masked_probas = numpy.asarray([numpy.multiply(proba_array[:, :, i], mask) for i in range(0, len(self.classes_))])
                p_sum = numpy.sum(masked_probas, axis=1)
                proba = (p_sum/numpy.sum(p_sum, axis=0)).T
        else:
            # Option 2: conf_thr is not None, neither n_decisors nor perc_decisors are set
            # Thus, base-learners contribute if they are confident at least 'conf_thr'
            conf_array = conf_array.transpose()
            mask = numpy.asarray(conf_array >= self.conf_thr).T
            masked_probas = numpy.asarray(
                [numpy.multiply(proba_array[:, :, i], mask) for i in range(0, len(self.classes_))])
            p_sum = numpy.sum(masked_probas, axis=1)
            proba = (p_sum / numpy.sum(p_sum, axis=0)).T

        # Final averaged Result
        if get_base:
            return proba, base_dict
        else:
            return proba

    def draw_samples(self, X, y, samples_n: int, weights = None):
        """
        Returns samples of a labeled set (X, y), y may be None
        :param weights: may be none, weights for sampling data (needed mostly for ConfBoost)
        :param X: the set to sample
        :param y: the labels to sample
        :param samples_n: numer of samples
        :return: a subset of (X, y)
        """
        if weights is None:
            indexes = numpy.random.choice(X.shape[0], samples_n, replace=False, p=None)
        else:
            indexes = numpy.random.choice(len(weights), samples_n, replace=False, p=weights)
        sample_x = numpy.asarray(X[indexes, :])
        # If data is labeled we also have to refine labels
        if y is not None and hasattr(self, 'classes_') and self.classes_ is not None and len(self.classes_) > 1:
            sample_y = y[indexes]
            sample_labels = unique_labels(sample_y)
            missing_labels = [item for item in self.classes_ if item not in sample_labels]
            # And make sure that there is at least a sample for each class of the problem
            if missing_labels is not None and len(missing_labels) > 0:
                # For each missing class
                for missing_class in missing_labels:
                    miss_class_indexes = numpy.asarray(numpy.where(y == missing_class)[0])
                    new_sampled_index = numpy.random.choice(miss_class_indexes, None, replace=False, p=None)
                    X_missing_class = X[new_sampled_index, :]
                    sample_x = numpy.append(sample_x, [X_missing_class], axis=0)
                    sample_y = numpy.append(sample_y, missing_class)
        else:
            sample_y = None
        return sample_x, sample_y

    def predict_confidence(self, X):
        """
        Method to compute the confidence in predictions of a classifier
        :param X: the test set
        :return: array of confidence scores
        """
        conf = numpy.zeros(X.shape[0])
        for clf in self.estimators_:
            conf += predict_confidence(clf, X)
        return conf / self.n_base

    def get_feature_importances(self):
        """
        Placeholder, to be implemented if possible
        :return: feature importances (to be tested)
        """
        fi = []
        for clf in self.estimators_:
            c_fi = clf.get_feature_importances()
            fi.append(c_fi)
        fi = numpy.asarray(fi)
        return numpy.average(fi, axis=1)

    def predict(self, X, get_base: bool = False):
        """
        Method to compute predict of a classifier
        :param get_base: True if predictions of base-learners have to be provided as well
        :param X: the test set
        :return: array of predicted class and (if get_base = True) dictionary with base-learners predictions
        """
        if get_base is True:
            proba, base_probas = predict_proba(self, X, get_base=True)
            pred_base = [self.classes_[numpy.argmax(base_p, axis=1)] for base_p in base_probas.values()]
            return self.classes_[numpy.argmax(proba, axis=1)], dict(zip(base_probas.keys(), pred_base))
        else:
            proba = predict_proba(self, X)
            return self.classes_[numpy.argmax(proba, axis=1)]

    def decision_function(self, X):
        """
        Compatibility with PYOD
        :param X: the test set
        :return: a numpy array, or None
        """
        if X is None:
            return None
        X = check_array(X)
        probas = predict_proba(self, X)
        if probas.shape[1] >= 2:
            return numpy.sum(probas[:, 1:], axis=1)
        else:
            return numpy.zeros(X.shape[0])

    def compute_feature_importances(self):
        """
        Outputs feature ranking in building a Classifier
        :return: ndarray containing feature ranks
        """
        if hasattr(self.clf, 'feature_importances_'):
            return self.clf.feature_importances_
        elif hasattr(self.clf, 'coef_'):
            return numpy.sum(numpy.absolute(self.clf.coef_), axis=0)
        return []

    def is_unsupervised(self):
        """
        true if the classifier is unsupervised
        :return: boolean
        """
        return hasattr(self, 'classes_') and numpy.array_equal(self.classes_, [0, 1])

    def classifier_name(self):
        """
        Returns the name of the classifier (as string)
        """
        return self.clf.__class__.__name__

    def get_diversity(self, X, y, metrics=None):
        """
        Returns diversity metrics. Works only with ensembles.
        :param metrics: name of the metrics to output (list of Metric objects)
        :param X: test set
        :param y: labels of the test set
        :return: diversity metrics
        """
        X = check_array(X)
        predictions = []
        check_is_fitted(self)
        ens_pred, predictions = self.predict(X, True)
        predictions = numpy.column_stack(list(predictions.values()))
        if predictions is not None and len(predictions) > 0:
            # Compute metrics
            metric_scores = {}
            if metrics is None or not isinstance(metrics, list):
                metrics = get_default()
            for metric in metrics:
                metric_scores[metric.get_name()] = metric.compute_diversity(predictions, y)
            return metric_scores
        else:
            # If it is not an ensemble
            return {}
